#uses "fwUnitTestComponentAsserts.ctl"

#uses "test/fwRPC/utils.ctl"

#uses "fwRPC/fwRPC_client.ctl"
#uses "fwRPC/fwRPC_server.ctl"
#uses "fwRPC/fwRPC_common.ctl"
#uses "fwRPC/fwRPC_pmon.ctl"
#uses "fwRPC/fwRPC_worker.ctl"
#uses "fwRPC/fwRPC_core.ctl"
#uses "fwRPC/fwRPC_RingBuffer.ctl"
#uses "fwRPC/fwRPC_msgBroker.ctl"


private global int _broker_index;
private global int _core_index;
private global int _worker_index;

private global string _broker_cmd = "fwRPC/fwRPC_broker_script.ctl";
private global string _core_cmd   = "fwRPC/fwRPC_core_script.ctl -1 -1 true testCore";
private global string _worker_cmd = "fwRPC/fwRPC_worker_script.ctl";

private global int _id_worker;
private global string _dpWorkerRb;

const string test_libname = "test/fwRPC/test_fwRPC_dist.ctl";

const string test_funcname = "func";
const string test_stubname = "mock_stub";
const string test_result = "testresult";

const string test_funcname_add = "func_add";
const string test_stubname_add = "mock_add";

const string test_funcname_err = "func_err";
const string test_stubname_err = "mock_err";

global private string _ctx;


// Setup
//------------------------------------------------------------------------------

void test_fwRPC_dist_setupSuite() {
  _BEGIN_SUITE("distributed integration");

  fwRPC_registerFunction(
    test_funcname,
    test_stubname,
    test_libname,
    -1,
    makeDynInt(ANYTYPE_VAR),
    true
  );

  fwRPC_registerFunction(
    test_funcname_add,
    test_stubname_add,
    test_libname,
    -1,
    makeDynInt(DYN_INT_VAR, DYN_FLOAT_VAR)
  );

  fwRPC_registerFunction(
    test_funcname_err,
    test_stubname_err,
    test_libname,
    -1,
    makeDynInt(ANYTYPE_VAR)
  );

  dyn_mixed coninfo = fwRPC_pmon_getConInfo();
  stopWorkers(coninfo);
  
  dyn_dyn_mixed managers;
  dyn_dyn_mixed managersStat;
  fwRPC_pmon_getManagerList(coninfo, managers);
  fwRPC_pmon_getManagerStati(coninfo, managersStat);
  for (int i=1; i<=dynlen(managers); ++i) {
    dyn_mixed manager = managers[i];
    if (dynlen(manager) < fwRPC_PMON_LIST_CMD) continue;

    if (
      manager[fwRPC_PMON_LIST_CMD] == "fwRPC/fwRPC_broker_script.ctl" ||
      strpos(manager[fwRPC_PMON_LIST_CMD], "fwRPC/fwRPC_core_script.ctl") >= 0 ||
      managers[fwRPC_PMON_LIST_CMD] == "fwRPC/fwRPC_worker_script.ctl"
    ) {
      if (managersStat[i][fwRPC_PMON_STAT_STATE] != fwRPC_PMON_MANINFO_STATE_RUNNING)
        continue;

      fwRPC_pmon_stopCtrlManager(coninfo, i-1);
    }
  }

  _broker_index = fwRPC_pmon_appendCtrlManager(coninfo, _broker_cmd);
  _core_index   = fwRPC_pmon_appendCtrlManager(coninfo, _core_cmd);
  _worker_index = fwRPC_pmon_appendCtrlManager(coninfo, _worker_cmd);

  fwRPC_pmon_startCtrlManager(coninfo, _broker_index);
  fwRPC_pmon_startCtrlManager(coninfo, _core_index);
  fwRPC_pmon_startCtrlManager(coninfo, _worker_index);

  string clientName = "testClient";
  fwRPC_client_startMonitor(_ctx, clientName, true);

  fwRPC_pmon_getWorkerCtrlManagers(coninfo, managers);
  _id_worker = managers[dynlen(managers)][fwRPC_PMON_MANINFO_NUM];
  _dpWorkerRb = fwRPC_common_makeDpName(
    fwRPC_RBMBX_DPT,
    fwRPC_DIST_TAG_LOCAL + "_" + fwRPC_DIST_TAG_WORKER + "_" + _id_worker
  );
}

void test_fwRPC_dist_teardownSuite() {
  fwRPC_client_stopMonitor(_ctx);

  dyn_mixed coninfo = fwRPC_pmon_getConInfo();

  dyn_dyn_mixed managers;
  fwRPC_pmon_getWorkerCtrlManagers(coninfo, managers);
  int num = dynlen(managers);

  DebugTN("Stopping CTRL managers..");
  fwRPC_pmon_stopCtrlManager(coninfo, _worker_index);
  fwRPC_pmon_stopCtrlManager(coninfo, _core_index);
  fwRPC_pmon_stopCtrlManager(coninfo, _broker_index);

  DebugTN("Removing CTRL managers..");
  fwRPC_pmon_deleteCtrlManager(coninfo, _worker_index);
  fwRPC_pmon_deleteCtrlManager(coninfo, _core_index);
  fwRPC_pmon_deleteCtrlManager(coninfo, _broker_index);

  clearFuncDp(test_funcname);

  _END_SUITE();
}

void test_fwRPC_dist_setup() {
  _BEGIN_TEST();
}

void test_fwRPC_dist_teardown() {

}



// Mock / helpers
//------------------------------------------------------------------------------

anytype mock_stub(anytype hdl, anytype params) {
  delay(0,500);
  return fwRPC_makeResult(hdl, true, test_result);
}

anytype mock_add(anytype hdl, anytype params) {
  return fwRPC_makeResult(hdl, true, params[1] + params[2]);
}

anytype mock_err(anytype hdl, anytype params) {
  throw(makeError("", PRIO_SEVERE, ERR_PARAM, 54, "exception text"));
}


// Tests
//------------------------------------------------------------------------------

void test_fwRPC_dist_listServices() {
  dyn_string services;
  int r = fwRPC_client_listServices(_ctx, services, true, 5000);

  assertEqual(0, r, "timeout");
  assert(dynlen(services) > 0);
  bool found = false;
  for (int i=1; i<=dynlen(services); i++) {
    if (services[i] == test_funcname) {
      found = true;
      break;
    }
  }
  assertTrue(found);
}

void test_fwRPC_dist_callSync() {
  anytype result;
  int r = fwRPC_client_callSync(_ctx, test_funcname, "", result);

  assertTrue(r != -1, "timeout");
  assertEqual(fwRPC_REQUEST_STATUS_DONE, r);
  assertEqual(test_result, result);
}

void test_fwRPC_dist_subrequest() {
  anytype hdl;

  fwRPC_client_call(_ctx, hdl, test_funcname, "");

  int status;
  anytype result;

  _waitForCompletion(_ctx, hdl, status, result);

  assertEqual(fwRPC_REQUEST_STATUS_DONE, status);
  assertEqual(test_result, result);
}


void test_fwRPC_dist_requestAndTimeout() {
  anytype hdl;
  fwRPC_client_call(_ctx, hdl, test_funcname, "");

  int status;
  anytype result;

  int ret = fwRPC_client_getResult(_ctx, hdl, result, true, 50);
  assertEqual(fwRPC_CLIENT_E_TIMEOUT, ret);
}

void test_fwRPC_dist_requestAndError() {
  anytype hdl;

  fwRPC_client_call(_ctx, hdl, test_funcname_err, "");

  int status;
  anytype result;
  int newStatus = _waitForCompletion(_ctx, hdl, status, result);

  assertEqual(fwRPC_REQUEST_STATUS_ERROR, newStatus, "timeout");
  assertEqual(fwRPC_REQUEST_STATUS_ERROR, status, "not error state");
  assertEqual(
    fwException_getClass(fwRPC_EX_WORKER_EXECUTION),
    fwException_getClass(result)
  );
}

void test_fwRPC_dist_cancel() {
  anytype hdl;

  fwRPC_client_call(_ctx, hdl, test_funcname, "");

  fwRPC_client_cancel(_ctx, hdl);

  int status; anytype result;
  _waitForCompletion(_ctx, hdl, status, result);

  assert(fwRPC_REQUEST_STATUS_CANCEL, status);
}

void test_fwRPC_dist_exclusiveExecution() {
  _startJobWatch(1, _dpWorkerRb);
  anytype hdl;

  fwRPC_client_call(_ctx, hdl, test_funcname, "");
  fwRPC_client_call(_ctx, hdl, test_funcname, "");
  fwRPC_client_call(_ctx, hdl, test_funcname, "");
  fwRPC_client_call(_ctx, hdl, test_funcname, "");

  int status;
  string result;
  _waitForCompletion(_ctx, hdl, status, result, 10000);

  _stopJobWatch(_dpWorkerRb);
}

void test_fwRPC_dist_requestSpecificType() {
  anytype hdl;

  fwRPC_client_call(_ctx, hdl, test_funcname_add, makeDynInt(1,4));

  int status;
  anytype result;
  _waitForCompletion(_ctx, hdl, status, result);

  assertEqual(fwRPC_REQUEST_STATUS_DONE, status);
  assertEqual(5, result);
}

void test_fwRPC_dist_requestSpecificTypeError() {
  anytype hdl;

  fwRPC_client_call(_ctx, hdl, test_funcname_add, "string");

  int status;
  anytype result;
  _waitForCompletion(_ctx, hdl, status, result);

  assertEqual(fwRPC_REQUEST_STATUS_ERROR, status);
  assertEqual(
    fwException_getClass(fwRPC_EX_INVALID_PARAM),
    fwException_getClass(result)
  );
}


void test_fwRPC_dist_broker_returns_error_result_for_nonexistant_func() {
  anytype hdl;

  fwRPC_client_call(_ctx, hdl, "idontexist", "");

  int status; anytype result;
  _waitForCompletion(_ctx, hdl, status, result);

  assertEqual(fwRPC_REQUEST_STATUS_ERROR, status);
  assertEqual(ERRCLASS_VAR, getType(result));
  assertEqual(
    fwException_getClass(fwRPC_EX_INVALID_PARAM),
    fwException_getClass(result)
  );
}

void test_fwRPC_dist_workerQueuesNonReady() {
  _startJobWatch(1, _dpWorkerRb);

  string workerProject = fwRPC_common_cleanProjectName(getSystemName());
  string workerName = "loc_work_" + _id_worker;

  dyn_mixed req;
  req[fwRPC_DISTMSG_REQUEST_HDL] = "1";
  req[fwRPC_DISTMSG_REQUEST_FUNC] = test_funcname;
  req[fwRPC_DISTMSG_REQUEST_PARAMS] = "";

  fwRPC_msgBroker_listen(true);

  fwRPC_msgBroker_writeMessage(
    workerProject, workerName,
    fwRPC_DISTMSG_TYPE_REQUEST, req
  );

  req[fwRPC_DISTMSG_REQUEST_HDL] = "2";
  fwRPC_msgBroker_writeMessage(
    workerProject, workerName,
    fwRPC_DISTMSG_TYPE_REQUEST, req
  );

  fwRPC_msgBroker_stop();

  _stopJobWatch(_dpWorkerRb);
}

void test_fwRPC_dist_recovery() {
  dyn_mixed coninfo = fwRPC_pmon_getConInfo();
  fwRPC_pmon_stopCtrlManager(coninfo, _worker_index);

  anytype hdl;

  fwRPC_client_call(_ctx, hdl, test_funcname, "");

  _subscribeHdl(_ctx, hdl);
  int ok = _waitForStatus(hdl, fwRPC_REQUEST_STATUS_QUEUED);
  assertTrue(ok);

  fwRPC_client_recover(_ctx, "mock_statusCB");

  _clearStatuses(hdl);

  ok = _waitForStatus(hdl, fwRPC_REQUEST_STATUS_QUEUED);
  assertTrue(ok);

  //wait for core to delete the DP before starting again
  while(dpExists(_dpWorkerRb)) {delay(0,500);}
  fwRPC_pmon_startCtrlManager(coninfo, _worker_index);
  while(!dpExists(_dpWorkerRb)) {delay(0,500);}

}


void test_fwRPC_dist_broker_distributes_core_loads() {
  //second local core should have the same function list, different RBMBX
  //note that two local cores are not typical in an fwRPC setup
  string corename = fwRPC_DIST_TAG_REMOTE + "_" + fwRPC_DIST_TAG_CORE + "_a";
  string rbmbx = fwRPC_common_makeDpName(fwRPC_RBMBX_DPT, corename);

  //create our dummy core rbmbx
  if (!dpExists(rbmbx)) dpCreate(rbmbx, fwRPC_RBMBX_DPT);
  fwRPC_RingBuffer_init(rbmbx + ".in", 50);
  fwRPC_RingBuffer_init(rbmbx + ".out", 50);

  //send a core register msg to the broker
  string msg = fwRPC_dist_buildMessage(
    fwRPC_DISTMSG_TYPE_REGISTER,
    fwRPC_DIST_BRKR, fwRPC_DIST_BRKR,
    test_funcname
  );

  fwRPC_RingBuffer_write(rbmbx + ".out", msg);
  delay(1);

  //make two calls, each should be attributed to a different core
  anytype hdl1, hdl2;
  fwRPC_client_call(_ctx, hdl1, test_funcname, "");
  fwRPC_client_call(_ctx, hdl2, test_funcname, "");

  dyn_string processedHdls;
  delay(2);

  string buffer;
  while(fwRPC_RingBuffer_read(rbmbx + ".in", buffer) == fwRPC_RingBuffer_ESUCCESS) {
    string id, type, proj, name; mixed payload;

    fwRPC_common_unpackMessage(buffer, id, type, proj, name, payload);
    deserializeFromXml(payload, payload);

    if (type == fwRPC_DISTMSG_TYPE_REQUEST) dynAppend(
      processedHdls,
      payload[fwRPC_DISTMSG_REQUEST_HDL]
    );
  }

  assertEqual(1, dynlen(processedHdls),
    "Second core should receive exactly one of the requests");

  int status; anytype result;
  anytype hdl = processedHdls[1] == hdl1 ? hdl2 : hdl1;
  int status = fwRPC_client_getResult(_ctx, hdl, result);
  assertFalse(0, status, "timeout");
  assertEqual(fwRPC_REQUEST_STATUS_DONE, status, "status should be done");

  msg = fwRPC_dist_buildMessage(
    fwRPC_DISTMSG_TYPE_DISCONNECT,
    fwRPC_DIST_BRKR, fwRPC_DIST_BRKR,
    ""
  );
  fwRPC_RingBuffer_write(rbmbx + ".out", msg);
  delay(1);

  //cleanup
  dpDelete(rbmbx);
}


void test_fwRPC_dist_broker_selects_new_core_if_call_status_not_received() {
  anytype hdl;
  dyn_mixed coninfo = fwRPC_pmon_getConInfo();
  string corename = fwRPC_DIST_TAG_REMOTE + "_" + fwRPC_DIST_TAG_CORE + "_testCore";
  string rbmbx = fwRPC_common_makeDpName(fwRPC_RBMBX_DPT, corename);

  delay(1);

  //core should still be registered at the broker after sudden ctrl mngr stop
  //NOTE: worker ringbuffer_read error is to be expected after this line
  fwRPC_pmon_stopCtrlManager(coninfo, _core_index);

  fwRPC_RingBuffer_init(rbmbx + ".in", 50);

  //broker will assign to previous core, and should time out
  fwRPC_client_call(_ctx, hdl, test_funcname, "12344");

  delay(1);

  string buffer, id, type, proj, name; mixed payload;

  int err = fwRPC_RingBuffer_read(rbmbx + ".in", buffer);
  assertEqual(fwRPC_RingBuffer_ESUCCESS, err, "core should receive req");

  fwRPC_common_unpackMessage(buffer, id, type, proj, name, payload);
  deserializeFromXml(payload, payload);

  assertEqual(fwRPC_DISTMSG_TYPE_REQUEST, type, "msg should be req");
  assertEqual(hdl, payload[fwRPC_DISTMSG_REQUEST_HDL], "req hdl should match");
  string prev_id = id;

  //broker will attempt to reassign to the next available core that can fulfill
  //the request. In this case it is the same as the previous core.

  delay(8);

  err = fwRPC_RingBuffer_read(rbmbx + ".in", buffer);
  assertEqual(fwRPC_RingBuffer_ESUCCESS, err, "core should receive new req");

  fwRPC_common_unpackMessage(buffer, id, type, proj, name, payload);
  deserializeFromXml(payload, payload);

  assertEqual(fwRPC_DISTMSG_TYPE_REQUEST, type, "new msg should be req");
  assertEqual(hdl, payload[fwRPC_DISTMSG_REQUEST_HDL], "new msg hdl should match");
  assert(prev_id != id, "msg ids should differ");

  //reset to previous state
  //NOTE: this will cause a lot of warning messages to show up in the debug panel
  //      about procedure re-registration at the broker and worker hb
  //      notifications at the core.
  fwRPC_pmon_startCtrlManager(coninfo, _core_index);
}
